
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/rumd_base.h"
#include "rumd/rumd_utils.h"
#include "rumd/PairPotential.h"
#include "rumd/NeighborList.h"

#include <fstream>
#include <typeinfo>
#include <sstream>
#include <iostream>

__device__ __constant__ float d_params[MaxNumPP_Objects][MaxNumTypes][MaxNumTypes][NumParam];
std::vector<unsigned int> PairPotential::objectIndexList;

// Forward declaration.
template<int STR, int CUTOFF, class P, class S> __device__ __host__ float fij( P* Pot, float4 my_r, float4 rj, float4* my_f, float4* my_w, 
									       float4* my_sts, float* param, S* simBox, float* simBoxPointer );


template<int STR, int CUTOFF, bool initialize, class P, class S> __device__ void calc_f( P *Pot, float4 my_r, float4 *r, float4 *f, float4 *w, float4 *sts, 
											 unsigned int my_type, S* simBox, float* simBoxPointer, NBTYPEv *NbList, 
											 unsigned int objectIdx );

template<int STR, int CUTOFF, bool initialize, class P, class S> __global__ void md_calc_f( P *Pot, float4* r, float4* f, float4* w, float4* sts, S* simBox, 
											    float* simBoxPointer, NBTYPEv* NbList, unsigned int objectIdx );

__device__ float4 Plus_Float4( const float4 &a, const float4 &b){ float4 c = {a.x + b.x, a.y + b.y, a.z + b.z, a.w + b.w}; return c; }

//////////////////////////////////////////////////
// Abstract PairPotential implementations.
//////////////////////////////////////////////////


PairPotential::PairPotential( CutoffMethod cutoff_method ) : neighborList(), testRSB(0), allocatedSize(0), cutoffMethod(cutoff_method), pp_objectIndex(GetNewObjectIndex()), d_f_pe(0), h_f_pe(0){ 
  memset(h_params, 0, MaxNumTypes * MaxNumTypes * NumParam * sizeof(float));
  
  if(cutoffMethod < NS || cutoffMethod > SF) throw RUMD_Error("PairPotential","PairPotential","Invalid cutoffMethod");
  
  if(pp_objectIndex >= MaxNumPP_Objects)
    throw RUMD_Error("PairPotential", "PairPotential", "Maximum number of pair potential objects has been reached.");
  
  SetID_String("base_pair_pot");
}


unsigned int PairPotential::GetNewObjectIndex() {
  unsigned int currentSize = objectIndexList.size();
  unsigned int smallest_not_there = currentSize;
  
  if(currentSize > 0 && currentSize != objectIndexList[currentSize-1] + 1) {
    unsigned int index = 0;
    while(index == objectIndexList[index])
      index++;
    smallest_not_there = index;
  }
  
  objectIndexList.push_back(smallest_not_there);
  sort(objectIndexList.begin(), objectIndexList.end());
  return smallest_not_there;
}



void PairPotential::RemoveObjectIndex(unsigned int index) {
  std::vector<unsigned int>::iterator found_index = std::find(objectIndexList.begin(), objectIndexList.end(), index);
  if(found_index != objectIndexList.end())
    objectIndexList.erase(found_index);
  else
    throw RUMD_Error("PairPotential","RemoveIndex","Could not find index for pair potential object");
}


void PairPotential::SetAllParams(float params[MaxNumTypes][MaxNumTypes][NumParam]) {
  unsigned int size = sizeof(float) * MaxNumTypes * MaxNumTypes * NumParam;
  memcpy(h_params, params, size);  
  neighborList.SetCutoffs(this);

  cudaMemcpyToSymbol(d_params, h_params, size, pp_objectIndex * size);
}

void PairPotential::CopyParamsToGPU( unsigned i, unsigned j, unsigned usedParams ){
  if( i >= MaxNumTypes || j >= MaxNumTypes )
    throw RUMD_Error("CopyParamsToGPU", GetID_String(), "Type index exceeds maximum number of types");  
  
  if( (usedParams + 2) > NumParam )
    throw( RUMD_Error("CopyParamsToGPU", GetID_String(), "RUMD requires more allocated parameters to run. Increase NumParam in rumd_base.h") );
  
  float4 my_f = { 0, 0, 0, 0 }; float4 my_w = { 0, 0, 0, 0 };
  
  float* param = h_params[i][j];
  float s = ComputeInteraction_CPU( param[0] * param[0], param, &my_f, &my_w );
  
  param[NumParam-1] = my_f.w;       // Potential at Rcut.
  param[NumParam-2] = s * param[0]; // Force at Rcut.
  
  // Copy pair-interaction parameters to Device
  if(verbose)
    std::cout << "Copying pair-interaction parameters to GPU (index=" << pp_objectIndex  << ", label=" << GetID_String() << ")" << std::endl;

  unsigned int size = sizeof(float) * MaxNumTypes * MaxNumTypes * NumParam;
  // need to use the "offset" argument to cudaMemcpyToSymbol ... doing
  // pointer arithmetic on the first argument doesn't work
  cudaMemcpyToSymbol(d_params, h_params, size, pp_objectIndex * size);
  // want to be sure that the neighborlist is correctly updated if the cutoff 
  // is changed
  neighborList.SetCutoffs(this);

} 

// Writes the actual potential and radial force implementation in PairPotential.h. 
// MaxNumTypes is used and should be changed to actual types.
void PairPotential::WritePotentials(SimulationBox* simBox){
  testRSB = dynamic_cast<RectangularSimulationBox*>(simBox);
  if(!testRSB)
    throw RUMD_Error("PairPotential","WritePotentials","Works only with RectangularSimulationBox for now");

  float4 my_r = {0.1f, 0.0f, 0.0f, 0.0f};
  float4    r = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_f = {0.0f, 0.0f, 0.0f, 0.0f}; 
  float4 my_w = {0.0f, 0.0f, 0.0f, 0.0f}; 
  float s = 0; float dist; 
  
  std::ofstream pFile( (std::string("potentials_") + GetID_String() + ".dat").c_str() );
  float halfBoxLength = powf( simBox->GetVolume(), 1.f/3.f ) / 2.0;
  
  while ( my_r.x < halfBoxLength ){
    pFile << my_r.x << " ";
    
    for(unsigned int MyType=0; MyType < MaxNumTypes; MyType++){
      for(unsigned int Type=0; Type < MaxNumTypes; Type++){
	s = 0; my_f.w = 0.f; 
	
	float* param = h_params[MyType][Type];
	float4 distance = testRSB->calculateDistance( my_r, r, testRSB->GetHostPointer() );
	
	if( distance.w <= (param[0]*param[0]) && distance.w >= 0.000001f ){
	  s = ComputeInteraction_CPU( distance.w, param, &my_f, &my_w );
	    
	  // Which cut-off method?
	  switch( cutoffMethod ){
	  case NS:
	    break;
	  case SP:
	    my_f.w -= param[NumParam-1]; 
	    break;
	  case SF:
	    dist = sqrtf(distance.w); 
	    s -= param[NumParam-2] / dist;  
	    my_f.w += param[NumParam-2] * ( dist - param[0] ) - param[NumParam-1]; 
	    my_w.w -= param[NumParam-2] * dist;
	    break;
	  default:
	    break;
	  }
	}
	pFile << my_f.w << " " << my_r.x * s << " ";
      }
    }
    pFile << std::endl;
    // Update only the x component.
    my_r.x += 0.001;
  } 
  pFile.close();
}

void PairPotential::AllocatePE_Array(unsigned int nvp) {
  if(allocatedSize != 0) {
    cudaFreeHost(h_f_pe);
    cudaFree(d_f_pe);
  }
  
  if( cudaMallocHost( (void**) &h_f_pe, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_f") );
  if( cudaMalloc( (void**) &d_f_pe, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on d_f") );
  memset( h_f_pe, 0, nvp * sizeof(float4) );
  cudaMemset( d_f_pe, 0, nvp * sizeof(float4) );
  
  allocatedSize = nvp;
}

double PairPotential::GetPotentialEnergy(){
  neighborList.UpdateNBlist(simBox);
  
  if(allocatedSize != particleData->GetNumberOfVirtualParticles())
    AllocatePE_Array(particleData->GetNumberOfVirtualParticles());
  
  // launch kernel with d_f_pe instead of df as the force array
  LaunchPE_Kernel();
  
  double pe_thisPotential = 0.;

  // copy to host
  cudaMemcpy( h_f_pe, d_f_pe, allocatedSize * sizeof(float4), cudaMemcpyDeviceToHost );

  // sum over particles
  for(unsigned int i=0; i < nParticles; i++)
    pe_thisPotential += h_f_pe[i].w;
  
  return pe_thisPotential;
}

void PairPotential::SetSampleParameters(KernelPlan& kPlan,  ParticleData* pd, SimulationBox* simBox, float4* d_misc){
  Potential::SetSampleParameters(kPlan, pd, simBox, d_misc);
  neighborList.Initialize(pd, kPlan);
  neighborList.SetCutoffs(this);
  testRSB = dynamic_cast<RectangularSimulationBox*>(simBox);
}

void PairPotential::SetExclusionBond(int1 *h_btlist, uint1* h_blist, unsigned num_part, unsigned num_virt_part, unsigned max_num_bonds, unsigned etype ){
  neighborList.SetExclusionBond(h_btlist, h_blist, num_part, num_virt_part, max_num_bonds, etype);
}

void PairPotential::SetExclusionType( unsigned type0, unsigned type1 ){
  neighborList.SetExclusionType( h_Type, type0, type1 );
}

void PairPotential::SetExclusionMolecule(int1 *h_mlist, unsigned molindex, unsigned max_num_uau){
  neighborList.SetExclusionMolecule(h_mlist, molindex, max_num_uau);
}

/////////////////////////////////////////////////////////
// Specialized PairPotentials implementations.
// The specialization is needed due to CUDA limitations.
/////////////////////////////////////////////////////////

void Pot_LJ_12_6::CalcF( bool initialize, bool calc_stresses ){
  neighborList.UpdateNBlist(simBox);

  if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
}

void Pot_LJ_12_6::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  }
}

void Pot_gLJ_m_n::CalcF(bool initialize, bool calc_stresses) {
  neighborList.UpdateNBlist(simBox);

  if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
}

void Pot_gLJ_m_n::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  }
}

void Pot_LJ_12_6_NEW::CalcF(bool initialize, bool calc_stresses) {
  neighborList.UpdateNBlist(simBox);
 
    if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
} 

void Pot_LJ_12_6_NEW::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  }
}

void Pot_LJ_12_6_Gauss::CalcF(bool initialize, bool calc_stresses){
  neighborList.UpdateNBlist(simBox);
  
    if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
} 

void Pot_LJ_12_6_Gauss::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  }
}

void  Pot_Gauss::CalcF(bool initialize, bool calc_stresses) {
  neighborList.UpdateNBlist(simBox);
  
    if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
}


void Pot_Gauss::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
  }
}

void Pot_Buckingham::CalcF(bool initialize, bool calc_stresses) {
  neighborList.UpdateNBlist(simBox);

    if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
}  

void Pot_Buckingham::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								  neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  }
}

void Pot_IPL_12::CalcF(bool initialize, bool calc_stresses) {
  neighborList.UpdateNBlist(simBox);
 
    if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
} 


void Pot_IPL_12::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  }
}

void Pot_IPL_18::CalcF(bool initialize, bool calc_stresses) {
  neighborList.UpdateNBlist(simBox);
  
    if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
} 

void Pot_IPL_18::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  }
}

void Pot_IPL_n::CalcF(bool initialize, bool calc_stresses) {
  neighborList.UpdateNBlist(simBox);

    if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
}  

void Pot_IPL_n::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  }
}

void Pot_Dzugutov::CalcF(bool initialize, bool calc_stresses) { 
  neighborList.UpdateNBlist(simBox);
   
    if( initialize ){
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
  else{
    if( calc_stresses ){
      switch( cutoffMethod ){
      case NS:
	md_calc_f<1,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<1,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<1,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
    else{
      switch( cutoffMethod ){
      case NS:
	md_calc_f<0,NS,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SP:
	md_calc_f<0,SP,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      case SF:
	md_calc_f<0,SF,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, testRSB, testRSB->GetDevicePointer(), 
								      neighborList.NbList, pp_objectIndex ); 
	break;
      }
    }
  }
}

void Pot_Dzugutov::LaunchPE_Kernel() {
  switch( cutoffMethod ){
  case NS:
    md_calc_f<0,NS,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SP:
    md_calc_f<0,SP,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  case SF:
    md_calc_f<0,SF,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, testRSB, testRSB->GetDevicePointer(),  
								 neighborList.NbList, pp_objectIndex );
    break;
  }
}

/////////////////////////////////////////////////////////
// Kernel implementation of pair force evaluation.
/////////////////////////////////////////////////////////

template<int STR, int CUTOFF, class P, class S> __device__ __host__ float fij( P* Pot, float4 my_r, float4 rj, float4* my_f, float4* my_w, 
									       float4* my_sts, float* param, S* simBox, float* simBoxPointer ){

  float4 dist = simBox->calculateDistance(my_r, rj, simBoxPointer);
  
  // Inside cut-off for interaction? (param[0]=Rcut)
  if ( dist.w <= (param[0]*param[0]) && dist.w >= 0.000001f ){
    float s = Pot->ComputeInteraction(dist.w, param, my_f, my_w); float r;
    
    // Which cut-off method?
    switch( CUTOFF ){
    case PairPotential::NS:
      break;
    case PairPotential::SP:
      (*my_f).w -= param[NumParam-1]; 
      break;
    case PairPotential::SF:
      r = sqrtf(dist.w); 
      s -= param[NumParam-2] / r; 
      (*my_f).w += param[NumParam-2] * ( r - param[0] ) - param[NumParam-1]; 
      (*my_w).w -= param[NumParam-2] * r; 
      break;
    default:
      break;
    }
    
    // F_ij.
    (*my_f).x += dist.x * s;
    (*my_f).y += dist.y * s;
    (*my_f).z += dist.z * s;
    
    // W_ij.
    (*my_w).w += dist.w * s;
    
    if(STR){
      // stress - diagonal components
      (*my_sts).x -= dist.x * dist.x * s;  // xx
      (*my_sts).y -= dist.y * dist.y * s;  // yy
      (*my_sts).z -= dist.z * dist.z * s;  // zz
      // stress - off-diagonal components
      (*my_sts).w -=  dist.y * dist.z * s; // yz
      (*my_w).y   -=  dist.x * dist.z * s; // xz
      (*my_w).z   -=  dist.x * dist.y * s; // xy      
    }
  } 
  return dist.w;                                                 
}

template<int STR, int CUTOFF, bool initialize, class P, class S> __device__ void calc_f( P* Pot, float4 my_r, float4* r, float4* f, float4* w, float4* sts, 
											 unsigned int my_type, S* simBox, float* simBoxPointer, 
											 NBTYPEv* NbList, unsigned int objectIdx ){
  extern __shared__ float4 s_r[];
  
  float4 total_f = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 total_w = {0.0f, 0.0f, 0.0f, 0.0f};  
  float4 total_sts = {0.0f, 0.0f, 0.0f, 0.0f};
  
  float4 my_f = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_w = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_sts = {0.0f, 0.0f, 0.0f, 0.0f};
  
  if(!initialize){
    total_w = w[MyGP];
    total_f = f[MyGP]; 
    
    if(STR)
      total_sts = sts[MyGP];
  }
  
  const unsigned int offset = MyT*PPerBlock;           
  const unsigned int CellOffset = MyB*NumBlocks*PPerBlock; 
  unsigned int Pread, j, cum_j;
  NBTYPE NbListCell;
  
  // Loop over all cells
  for( unsigned int cell0 = 0; cell0 < NumBlocks; cell0+=TPerPart ){
    unsigned int cell = cell0 + MyT;
    
    Pread = __umul24(cell,PPerBlock) + MyP;
    
    // Read coordinates from global to shared mem. NOTE: must be same method as in nb-list!!!
    if ( cell < NumBlocks ){
      s_r[MyP+offset] = r[Pread]; 
    }
    
    __syncthreads();  
    
    if (cell<NumBlocks){
      // Read the NbList (coded as a binary mask) from global memory to register 
      NbListCell = NbList[CellOffset + Pread].x;
      
      // Run through NbList (i.e. the '1'-bits in NbListCell)   
      cum_j = 0; 
      
      while (NbListCell>0) {
	cum_j += (j = __ffs(NbListCell));         // find next '1'-bit (from the right, labeling bit 0 as 1) 
	NbListCell = (NbListCell >> 1) >> j-1;    // remove the '1'-bit we just found
	
	int index = offset + cum_j - 1; 
	
	fij<STR, CUTOFF>( Pot, my_r, s_r[index], &my_f, &my_w, &my_sts, d_params[objectIdx][my_type][__float_as_int(s_r[index].w)], simBox, simBoxPointer );
      }  
    } // END if (cell<NumBlocks)
    __syncthreads();  
    
  } // END for( unsigned int cell0 = 0; ... 
  
  // Sum the force for all local threads. 'Borrowing s_r'
  s_r[MyP+offset] = my_f;
  
  __syncthreads();  
  
  if( MyT == 0 ){
    for( int i=1; i < TPerPart; i++ ){
      int j = MyP + __umul24(i,PPerBlock); 
      my_f.x += s_r[j].x;
      my_f.y += s_r[j].y;
      my_f.z += s_r[j].z;
      my_f.w += s_r[j].w;
    }

    my_f.w *= 0.5f;
    
    if(initialize)
      f[MyGP] = my_f;
    else
      f[MyGP] = Plus_Float4(my_f, total_f); // do the addition locally before writing
  }
  
  __syncthreads();  
  // Sum the atomic/molecular virial for all local threads.
  s_r[MyP+offset] = my_w;
  __syncthreads();  
  
  if( MyT == 0 ){
    for ( int i=1; i < TPerPart; i++ ){
      int j = MyP + __umul24(i,PPerBlock); 
      my_w.w += s_r[j].w;
      
      if(STR){
	my_w.y += s_r[j].y;
	my_w.z += s_r[j].z;
      }
    }

    if(initialize)
      w[MyGP] = my_w;
    else
      w[MyGP] = Plus_Float4(my_w, total_w);
  }
  
  if(STR){
    __syncthreads();  
    // Sum the shear stress for all local threads.
    s_r[MyP+offset] = my_sts;
    __syncthreads();  
    
    if( MyT == 0 ){
      for ( int i=1; i < TPerPart; i++ ){
	int j = MyP + __umul24(i,PPerBlock); 
	my_sts.x += s_r[j].x;
	my_sts.y += s_r[j].y;
	my_sts.z += s_r[j].z;
	my_sts.w += s_r[j].w;
      }
      if(initialize)
	sts[MyGP] = my_sts;
      else
	sts[MyGP] = Plus_Float4(my_sts, total_sts);
    }
  } // END if(STR)
} // END calc_f

template<int STR, int CUTOFF, bool initialize, class P, class S> __global__ void md_calc_f( P* Pot, float4* r, float4* f, float4* w, float4* sts, S* simBox, 
											    float* simBoxPointer, NBTYPEv* NbList, unsigned int objectIdx ){
  extern __shared__ float4 s_r[];
  
  // Load the simulation box in local memory to avoid bank conflicts.
  float array[simulationBoxSize];
  simBox->loadMemory( array, simBoxPointer );
  
  // One thread per particle reads the position
  if(MyT == 0){
    s_r[MyP] = r[MyGP];
  }

  __syncthreads();
  
  float4 my_r = s_r[MyP]; 
  
  __syncthreads(); 

  // Calculate and save force on this particle
  calc_f<STR, CUTOFF, initialize>( Pot, my_r, r, f, w, sts, __float_as_int(my_r.w), simBox, array, NbList, objectIdx );
}
