/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd/Sample.h"
#include "rumd/Integrator.h"
#include "rumd/IntegratorMetaData.h"
#include "rumd/Potential.h"
#include "rumd/PairPotential.h"
#include "rumd/ConstraintPotential.h"
#include "rumd/ConfigurationOutputManager.h"
#include "rumd/RUMD_Error.h"

#include <iostream>
#include <algorithm>

/////////////////////////////////////////////
// CONSTRUCTION/DESTRUCTION 
/////////////////////////////////////////////

Sample::Sample( unsigned int pb, unsigned int tp ) 
  : particleData(),
    trajectoryDir("./TrajectoryFiles"),
    blockSize(0),
    itg(0),
    itg_metaData(0),
    verbose(true)
{
  if( !( (pb != 0) && (pb <= 32) && ( (pb & (~pb + 1)) == pb ) ) )
    throw(RUMD_Error("Sample","Sample","pb is not a power of 2 or is larger than 32"));
  
  if( !( (tp != 0) && (tp <= 8) && ( (tp & (~tp + 1)) == tp ) ) )
    throw(RUMD_Error("Sample","Sample","tp is not a power of 2 or is larger than 8"));
  
  simulationBox = new RectangularSimulationBox();
  threads_per_particle = tp;
  particles_per_block = pb;
  degreesOfFreedom = 0;
  CreateDefaultOutputManagers();
}

Sample::~Sample(){
  delete restartWriter;
  delete restartLogLin;
  
  delete trajectoryWriter;
  delete trajectoryLogLin;

  delete energiesLogLin;

  delete simulationBox;
  if(itg_metaData)
    delete itg_metaData;
}


// Copy-constructor
Sample::Sample(const Sample& S) : particleData(),
				  trajectoryDir("./TrajectoryFiles"),
				  blockSize(0),
				  itg(0),
				  itg_metaData(0),
				  verbose(true) {

  threads_per_particle = S.GetThreadsPerParticle();
  particles_per_block = S.GetParticlesPerBlock();  
  simulationBox = S.GetSimulationBox()->MakeCopy();

  SetNumberOfParticles(S.GetNumberOfParticles());

  numTypes = S.GetNumberOfTypes();
  for(unsigned int type = 0; type < numTypes; type++)
    massOfType[type] = S.GetMass(type);

  particleData.CopyHostData(S.GetParticleData()); // copies only on CPU
  CopySimulationDataToDevice();

  // may want to make it optional to do the following, or have a different
  // trajectoryDir
  CreateDefaultOutputManagers();
}

Sample& Sample::operator=(const Sample& S){
  if(this != &S){ 
    if(GetNumberOfParticles() != S.GetNumberOfParticles())
      SetNumberOfParticles(S.GetNumberOfParticles());
    
    particleData = S.GetParticleData();

    simulationBox->CopyAnotherBox(S.GetSimulationBox());

    // Trigger neighborList update in any pair potential(s).
    std::vector<Potential*>::iterator potIter;
    for(potIter = potentialList.begin(); potIter != potentialList.end(); potIter++) {
      PairPotential* pair_pot = dynamic_cast<PairPotential*>(*potIter);
      if(pair_pot)
	pair_pot->ResetNeighborList();
    }
  }
  return *this; 
}

/////////////////////////////////////////////
// Initialization of Sample 
/////////////////////////////////////////////

void Sample::CreateDefaultOutputManagers() {
  writers_initialized = false;
  
  restartWriter = new ConfigurationWriterReader();
  restartWriter->metaData.precision = 9;
  restartWriter->metaData.has_images = true;
  restartWriter->metaData.has_velocities = true;
  restartWriter->metaData.has_forces = true;
  restartWriter->metaData.has_pot_energies = true;
  restartWriter->metaData.has_virials = true;
  
  restartLogLin = new ConfigurationOutputManager(this, trajectoryDir, "restart", restartWriter);
  restartLogLin->SetCreatesDirectory(true);
  restartLogLin->SetVerbose(verbose);
  outputManagers.push_back(restartLogLin);

  trajectoryWriter = new ConfigurationWriterReader();
  trajectoryWriter->metaData.precision = 4;
  trajectoryWriter->metaData.has_images = true;
  trajectoryWriter->metaData.has_velocities = false;
  trajectoryWriter->metaData.has_forces = false;
  trajectoryWriter->metaData.has_pot_energies = true;
  trajectoryWriter->metaData.has_virials = true;
  trajectoryWriter->metaData.has_logLin = true;  

  trajectoryLogLin = new ConfigurationOutputManager(this, trajectoryDir, "block", trajectoryWriter);
  trajectoryLogLin->SetDuplicateBlockEnds(true);
  trajectoryLogLin->SetVerbose(verbose);
  // make the default behaviour pure logarithmic saving with base 1:  
  trajectoryLogLin->SetLogLinParams(1,0);
  outputManagers.push_back(trajectoryLogLin);

  energiesLogLin = new EnergiesOutputManager(this, trajectoryDir, "energies");
  energiesLogLin->metaData.precision = 6;
  energiesLogLin->SetVerbose(verbose);
  outputManagers.push_back(energiesLogLin);
}

void Sample::SetNumberOfParticles(unsigned int set_num_part){
  particleData.SetNumberOfParticles(set_num_part, particles_per_block, threads_per_particle);
  degreesOfFreedom = 3*particleData.GetNumberOfParticles() - 3; 

  kPlan.grid.x = particleData.GetNumberOfBlocks();
  kPlan.threads.x = particles_per_block;
  kPlan.threads.y = threads_per_particle;
  kPlan.shared_size = particles_per_block * threads_per_particle * sizeof(float4);
  kPlan.num_blocks = particleData.GetNumberOfBlocks();
  kPlan.num_virt_part = particleData.GetNumberOfVirtualParticles();
  
  for(std::vector<Potential*>::iterator potIter = potentialList.begin(); potIter != potentialList.end(); potIter++)
    (*potIter)->SetSampleParameters(kPlan, &particleData, simulationBox, particleData.d_misc);
  
  if(itg)
    itg->SetSampleData(this, &particleData);
}

void Sample::SetPotential(Potential* potential){ 
  potentialList.clear();
  AddPotential(potential);
}

void Sample::AddPotential(Potential* potential){
  potential->SetVerbose(verbose);

  if(GetNumberOfParticles())
    potential->SetSampleParameters(kPlan, &particleData, simulationBox, particleData.d_misc);

  if(dynamic_cast<ConstraintPotential*>(potential)){
    potentialList.push_back(potential);
  }
  else{
    // Will induce reallocation, but this is a small list.
    potentialList.insert(potentialList.begin(), potential);
  }
  energiesLogLin->metaData.options[potential->GetID_String()] = false;
  energiesLogLin->metaData.fileStr[potential->GetID_String()] = potential->GetID_String();
}

// Set the current integrator.
void Sample::SetIntegrator(Integrator* newItg){
  itg = newItg;
  
  // SetNumberOfParticles has been called earlier and did not allocate.
  if(particleData.GetNumberOfParticles())
    itg->SetSampleData(this, &particleData);
}

void Sample::SetIntegratorMetaData(IntegratorMetaData* itg_md) {
  if(itg_metaData)
    delete itg_metaData;
  itg_metaData = itg_md;
}


///////////////////////////////////////////////////////////////
// Force and potential
///////////////////////////////////////////////////////////////

void Sample::CalcF(bool calc_stresses){
  if(!simulationBox)
    throw RUMD_Error("Sample","CalcF","No simulation box has been set");
  if(!potentialList.size())
    throw RUMD_Error("Sample","CalcF","No potential has been set");

  potentialList[0]->CalcF(true, calc_stresses);

  for(unsigned int pdx=1; pdx < potentialList.size(); ++pdx)
    potentialList[pdx]->CalcF(false, calc_stresses);

  if(itg)
    itg->CalculateAfterForce();
}  

double Sample::GetPotentialEnergy() const {
  particleData.CopyForFromDevice();
  double potentialEnergy = 0.;
  unsigned int nParticles = GetNumberOfParticles();

  for(unsigned int i=0; i < nParticles; i++)
    potentialEnergy += particleData.h_f[i].w;

  return potentialEnergy;
}

double Sample::GetVirial() const {
  particleData.CopyVirFromDevice();
  double virial = 0.;
  unsigned int nParticles = GetNumberOfParticles();

  for(unsigned int i=0; i < nParticles; i++) { 
    virial += particleData.h_w[i].w;
  }
  return virial/6.0;
}

////////////////////////////////////////////////////////
// IO   
////////////////////////////////////////////////////////

void Sample::SetVerbose(bool vb) {
  verbose = vb;
  particleData.SetVerbose(vb);
  std::vector<LogLinOutputManager*>::iterator llIter;
  for(llIter = outputManagers.begin(); llIter != outputManagers.end(); llIter++)
    (*llIter)->SetVerbose(verbose);

  for(unsigned int pdx=0; pdx < potentialList.size(); ++pdx)
    potentialList[pdx]->SetVerbose(verbose);

}


void Sample::SetOutputDirectory(const std::string& trajectoryDir) {
  this->trajectoryDir = trajectoryDir;
  std::vector<LogLinOutputManager*>::iterator llIter;
  for(llIter = outputManagers.begin(); llIter != outputManagers.end(); llIter++)
    (*llIter)->SetOutputDirectory(trajectoryDir);
  
  writers_initialized = false;
}


void Sample::SetOutputBlockSize(unsigned long int blockSize) {
  if(!blockSize) throw RUMD_Error("Sample","SetOutputBlockSize","blockSize must be greater than zero");
  this->blockSize = blockSize;
  
  std::vector<LogLinOutputManager*>::iterator llIter;
  for(llIter = outputManagers.begin(); llIter != outputManagers.end(); llIter++)
    (*llIter)->SetBlockSize(blockSize);

  restartLogLin->SetLogLinParams(blockSize,0);

  writers_initialized = false;
}

void Sample::SetWriteEnergies(bool writeEnergies) {
  std::vector<LogLinOutputManager*>::iterator ellIt = std::find(outputManagers.begin(), outputManagers.end(), energiesLogLin);
  bool energiesOnList = ( ellIt != outputManagers.end() );

  if(writeEnergies) {
    energiesLogLin->SetVerbose(verbose);
    energiesLogLin->SetBlockSize(blockSize);
  }

  if(writeEnergies && !energiesOnList)
    outputManagers.push_back(energiesLogLin);
  else if(!writeEnergies && energiesOnList)
    outputManagers.erase(ellIt);
    
}

void Sample::SetWriteTrajectory(bool writeTrajectory) {
  std::vector<LogLinOutputManager*>::iterator ellIt = std::find(outputManagers.begin(), outputManagers.end(), trajectoryLogLin);
  bool trajOnList = ( ellIt != outputManagers.end() );

  if(writeTrajectory) {
    trajectoryLogLin->SetVerbose(verbose);
    trajectoryLogLin->SetBlockSize(blockSize);
  }
  
  if(writeTrajectory && !trajOnList)
    outputManagers.push_back(trajectoryLogLin);
  else if(!writeTrajectory && trajOnList)
    outputManagers.erase(ellIt);
}

void Sample::SetWriteRestart(bool writeRestart) {
  std::vector<LogLinOutputManager*>::iterator ellIt = std::find(outputManagers.begin(), outputManagers.end(), restartLogLin);
  bool restartOnList = ( ellIt != outputManagers.end() );

  if(writeRestart) {
    restartLogLin->SetVerbose(verbose);
    restartLogLin->SetBlockSize(blockSize);
  }
  
  if(writeRestart && !restartOnList)
    outputManagers.push_back(restartLogLin);
  else if(!writeRestart && restartOnList)
    outputManagers.erase(ellIt);
}



void Sample::AddOutputManager(LogLinOutputManager* om) {
  om->SetVerbose(verbose);
  om->SetBlockSize(blockSize);
  outputManagers.push_back(om);
}

void Sample::SetTrajectoryLinearWriteInterval(unsigned long int writeInterval) { 
  trajectoryLogLin->SetLogLinParams(writeInterval, writeInterval);
  trajectoryLogLin->SetDuplicateBlockEnds(false);
  writers_initialized = false;
}


void Sample::SetTrajectoryLogLinParams(unsigned int base, unsigned int maxInterval) {
  trajectoryLogLin->SetLogLinParams(base, maxInterval);
  trajectoryLogLin->SetDuplicateBlockEnds(true);
  writers_initialized = false;
}


void Sample::SetEnergyLinearWriteInterval(unsigned int writeInterval) {
  energiesLogLin->SetLogLinParams(writeInterval, writeInterval);
  energiesLogLin->metaData.writeInterval = writeInterval;
  energiesLogLin->metaData.has_logLin = false;
  energiesLogLin->SetDuplicateBlockEnds(false); 
  writers_initialized = false;
}

void Sample::SetEnergyLogLinParams(unsigned int base, unsigned int maxInterval) {
  energiesLogLin->SetLogLinParams(base, maxInterval);
  energiesLogLin->metaData.writeInterval = 0.;
  energiesLogLin->metaData.has_logLin = true;
  energiesLogLin->SetDuplicateBlockEnds(true); 
  writers_initialized = false;
}

void Sample::InitializeOutputManagers(unsigned long int timeStepIndex) {
  // iterate over vector calling Initialize on each output manager 
  std::vector<LogLinOutputManager*>::iterator fwIter;
  for(fwIter = outputManagers.begin(); fwIter != outputManagers.end(); fwIter++)
    (*fwIter)->Initialize(timeStepIndex);
  writers_initialized = true;
}

void Sample::NotifyOutputManagers(unsigned long int timeStepIndex) {
  if(!writers_initialized)
    InitializeOutputManagers(timeStepIndex);
  
  // iterate over vector calling Update 
  std::vector<LogLinOutputManager*>::iterator fwIter;
  for(fwIter = outputManagers.begin(); fwIter != outputManagers.end(); fwIter++)
    (*fwIter)->Update(timeStepIndex);

}

void Sample::TerminateOutputManagers() {
  if(!writers_initialized)
    std::cerr << "Warning: TerminateOutputManagers called although they have not been initialized" << std::endl;
  
  // iterate over vector calling Update 
  std::vector<LogLinOutputManager*>::iterator llIter;
  for(llIter = outputManagers.begin(); llIter != outputManagers.end(); llIter++)
    (*llIter)->Terminate();

}

void Sample::ReadRestartConf( int restartBlock, unsigned int numDigitsBlkIdx ) {
  // following works only on CUDA 2.2 or higher
  std::ostringstream restartFilePath;
  restartFilePath << trajectoryDir << "/restart" << std::setfill('0') 
		  << std::setw(numDigitsBlkIdx) << std::setprecision(numDigitsBlkIdx)  
		  << restartBlock << ".xyz.gz";
  
  ReadConf(restartFilePath.str());
}

void Sample::ReadConf( const std::string& filename ) {
  ConfigurationWriterReader reader;
  reader.Read(this, filename);
}

//////////////////////////////////////////////
// Simulation modification. !! Adjust for topology file. !!
//////////////////////////////////////////////

// Calculate the CM of a molecule. The first particle defines the molecule.
float4 Sample::CalculateCenterOfMass( float4* position, float4* velocity, int moleculeIndex ){
  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(simulationBox);
  
  float moleculeMass = 1.f / velocity[moleculeIndex].w;
  float4 firstParticle = position[moleculeIndex];
  float4 CM = { moleculeMass * firstParticle.x, moleculeMass * firstParticle.y, moleculeMass * firstParticle.z, 0 }; 
  
  for(int i = moleculeIndex+1; i < moleculeIndex + 2; i++){ 
    float particleMass = 1.f / velocity[i].w;
    
    float4 difference = testRSB->calculateDistance( position[i], firstParticle, testRSB->GetHostPointer() );
    float3 nextParticle = { firstParticle.x + difference.x, firstParticle.y + difference.y, firstParticle.z + difference.z };
    
    CM.x += particleMass * nextParticle.x;
    CM.y += particleMass * nextParticle.y;
    CM.z += particleMass * nextParticle.z;
    moleculeMass += particleMass; 
  }
  
  CM.x /= moleculeMass;
  CM.y /= moleculeMass;
  CM.z /= moleculeMass;
  CM.w = ( 1.0 / moleculeMass );
  
  return CM;
}

// Simple particle scaling. Adjust for topology file.
void Sample::IsotropicScaleSystem( float Rscal ){

  particleData.CopyPosFromDevice();

  for(unsigned int i=0; i < particleData.GetNumberOfParticles(); i++){
    particleData.h_r[i].x *= Rscal;
    particleData.h_r[i].y *= Rscal;
    particleData.h_r[i].z *= Rscal;
  }
  
  simulationBox->ScaleBox( Rscal );  
  particleData.CopyPosToDevice();

}

//////////////////////////////////////////////
// Copy Function
//////////////////////////////////////////////

void Sample::CopySimulationDataToDevice(){
  particleData.CopyConfToDevice();
}

void Sample::CopySimulationDataFromDevice(){
  particleData.CopyConfFromDevice();
}
