
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/rumd_algorithms.h"
#include "rumd/rumd_technical.h"
#include <sstream>

__host__ __device__ float4 operator+ (const float4 &a, const float4 &b){
  float4 r;
  r.x = a.x + b.x;
  r.y = a.y + b.y;
  r.z = a.z + b.z;
  r.w = a.w + b.w;
  return r;
}

__host__ __device__ void operator+= (float4 &a, const float4 &b){
  a.x += b.x;
  a.y += b.y;
  a.z += b.z;
  a.w += b.w;
}

__host__ __device__ void operator+= (volatile float4 &a, volatile const float4 &b){
  a.x += b.x;
  a.y += b.y;
  a.z += b.z;
  a.w += b.w;
}

template <class T> std::string toString( T t, std::ios_base& (*f)(std::ios_base&) ){
  std::ostringstream oss;
  oss << f << t;
  return oss.str();
}

//////////////////////////////////////////////////////////
// Reduction algorithm: [Mark Harris, NVIDIA, SDK]
///////////////////////////////////////////////////////////

// Forward declaration.
template <class T, unsigned int blockSize> __global__ void reduce( T* g_idata, T* g_odata, unsigned int numParticles );

// Dynamically allocated shared memory is declared extern. Thus we need to define
// different names corresponding to different variable types.
template <typename T> struct SharedMemory{
  __device__ T* getPointer(){
    extern __device__ void error(void); // Undefined. Ensures that we won't compile any un-specialized types.
    error();
    return NULL;
  }
  __device__ T getZero(){ return NULL; }
};

template <> struct SharedMemory <float>{
  __device__ float* getPointer(){ extern __shared__ float s_float[]; return s_float; }    
  __device__ float getZero(){ return 0.f; }
};

template <> struct SharedMemory <float4>{
  __device__ float4* getPointer(){ extern __shared__ float4 s_float4[]; return s_float4; }    
  __device__ float4 getZero(){ float4 zero = { 0.f, 0.f, 0.f, 0.f }; return zero; }
};

inline void getThreadsAndBlocks( unsigned int size, unsigned int& numThreads, unsigned int& numBlocks, unsigned int maxThreadsPerBlock ){
  if(size == 1) 
    numThreads = 1;
  else
    numThreads = (size < 2 * maxThreadsPerBlock) ? size / 2 : maxThreadsPerBlock;
  
  numBlocks = size / (numThreads * 2);
}

template <class T> void sumIdenticalArrays( T* array, unsigned int numParticles, unsigned int numArrays, unsigned int maxThreadsPerBlock ){
  unsigned int numThreads = 0; unsigned int numBlocks = 0;
  unsigned int s = (unsigned) pow( 2, ceil( log( numParticles ) / log( 2 ) ) ); // Largest nearest power of 2.
  
  while(s > 1){
    getThreadsAndBlocks( s , numThreads, numBlocks, maxThreadsPerBlock );  
    dim3 dimGrid(numBlocks, numArrays);
    dim3 dimBlock(numThreads, 1);  
    
    // We must allocate 2 * numThreads * sizeof(T) due to warp unrolled in the reduction algorithm.
    unsigned int size = (numThreads <= 32) ? 2 * numThreads * sizeof(T) : numThreads * sizeof(T); 
    
    switch(numThreads){
    case 512:
      reduce<T,512><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 256:
      reduce<T,256><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 128:
      reduce<T,128><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 64:
      reduce<T,64><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 32:
      reduce<T,32><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 16:
      reduce<T,16><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 8:
      reduce<T,8><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 4:
      reduce<T,4><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 2:
      reduce<T,2><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 1:
      reduce<T,1><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    }
    
    s = s / (numThreads*2);
  }
}

// Performs a reduction.
template <class T, unsigned int blockSize> __global__ void reduce( T* g_idata, T* g_odata, unsigned int numParticles ){
  
  // Shared mem size is determined by the host app at run time.
  SharedMemory<T> smem;
  T* sdata = smem.getPointer();
  
  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x * (blockDim.x*2) + threadIdx.x;
  unsigned int offset = numParticles * blockIdx.y;
  
  T mySum = (i < numParticles) ? g_idata[offset+i] : smem.getZero();
  
  if( i + blockSize < numParticles ) 
    mySum += g_idata[offset+i+blockSize];  

  sdata[tid] = mySum;
  
  __syncthreads();
  
  // Unroll the initial iterations. We have to synchronize, since only 32 threads are atomic.
  if(blockSize >= 512) { if(tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
  if(blockSize >= 256) { if(tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
  if(blockSize >= 128) { if(tid <  64) { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }
  
  if(tid < 32){
    volatile T* localSdata = sdata; // We must declare volatile amongst the 32 warp threads.
    
    // Unroll the last warp.
    if(blockSize >= 64) localSdata[tid] += localSdata[tid + 32]; 
    if(blockSize >= 32) localSdata[tid] += localSdata[tid + 16]; 
    if(blockSize >= 16) localSdata[tid] += localSdata[tid +  8]; 
    if(blockSize >= 8)  localSdata[tid] += localSdata[tid +  4]; 
    if(blockSize >= 4)  localSdata[tid] += localSdata[tid +  2]; 
    if(blockSize >= 2)  localSdata[tid] += localSdata[tid +  1]; 
  }
    
  // Write result for this block to global mem 
  if(tid == 0) 
    g_odata[offset+blockIdx.x] = sdata[0];
}

/////////////////////////////////////////////////////////////////////
// PRNG algorithm: [L. Howes and D. Thomas, GPU Gems 3, Chapter 37]
/////////////////////////////////////////////////////////////////////

__host__ __device__ inline unsigned TausStep( unsigned &z, int S1, int S2, int S3, unsigned M ){
  unsigned b = ( ( ( z << S1 ) ^ z ) >> S2 );
  return z = ( ( ( z & M ) << S3 ) ^ b );
}

__host__ __device__ inline unsigned LCGStep( unsigned &z, unsigned A, unsigned C ){
  return z = ( A * z + C );
}

// Different prefactor will produce (0, 1].
__host__ __device__ inline float HybridTaus( uint4* state ){
  return 2.3283064365387e-10f* (
				TausStep(state->x, 13, 19, 12, 4294967294UL) ^
				TausStep(state->y,  2, 25,  4, 4294967288UL) ^
				TausStep(state->z,  3, 11, 17, 4294967280UL) ^
				LCGStep( state->w,  1664525,   1013904223UL)
				);
}

/////////////////////////////////////////////////////////////////////
// Box Muller Transformation.
/////////////////////////////////////////////////////////////////////

__host__ __device__ inline float2 BoxMuller( float u0, float u1 ){
  float r = sqrtf( -2.0f * logf(u0) );
  float theta = 2.f * float(M_PI) * u1;
  float2 value = { r*sinf(theta), r*cosf(theta) };
  return value;
}

//////////////////////////////////////////////////////////////////////
// Gaussian Elimination with rowise (partial) pivoting and 
// backsubstitution
//////////////////////////////////////////////////////////////////////

template <class T> __global__ void GaussPivotBack( T* x, T* b, T* A, unsigned* dimlist ){
  unsigned i=0, j=0, k, kk; 
  T tmp, c;
  
  unsigned offset = 0;
  for ( k=0; k<blockIdx.x; k++ ) offset += dimlist[k]*dimlist[k];
  T* Ap = &A[offset];
  
  offset = 0;
  for ( k=0; k<blockIdx.x; k++ ) offset += dimlist[k];
  
  T *bp = &x[offset];
  
  unsigned nrc  = dimlist[blockIdx.x];
  while ( i<nrc && j<nrc ){
    
    // Find pivot
    unsigned ir = i;
    for ( k=i+1; k<nrc; k++ ){
      if ( fabs(Ap[k*nrc+j]) > fabs(Ap[ir*nrc+j]) )
	ir = k;
    }
    
    if ( fabs(Ap[ir*nrc+j]) > FLT_EPSILON ){
      
      // Swap rows
      for ( k=0; k<nrc; k++ ){
	tmp = Ap[i*nrc+k];
	Ap[i*nrc+k] = Ap[ir*nrc+k];
	Ap[ir*nrc+k] = tmp;
      }
      
      tmp = bp[i];
      bp[i] = bp[ir];
      bp[ir] = tmp;
      
      // Divide row i (Eq. (i)) with Ap(i,j)
      c = Ap[i*nrc+j];
      bp[i] /= c;
      for ( k=0; k<nrc; k++ ) Ap[i*nrc+k] /= c;
      
      // Multiply and substract Ap(i,j) - Gaussian part
      for ( k=i+1; k<nrc; k++ ){
	c = Ap[k*nrc+j];
	for ( kk=0; kk<3; kk++ )
	  Ap[k*nrc+kk] -= c*Ap[i*nrc+kk];
	bp[k] -= c*bp[i];
      }
      
      i++;
    }
    
    j++;
  }
  
  // Backsubstitution
  for ( int i=nrc-1; i>=0; i-- ){
    
    float s = 0.0f;
    for ( int j=i+1; j<(signed)nrc; j++ )
      s += Ap[i*nrc+j]*bp[j];
    
    bp[i] = (bp[i] - s)/Ap[i*nrc+i];
  }
}

template <class T> void solveLinearSystems( T* x, T* b, T* A, unsigned* dimLinearSystems, unsigned nLinearSystems ){
  GaussPivotBack<<<nLinearSystems, 1>>>( x, b, A, dimLinearSystems );
}
